/**
Action Script library. String class
Copyright (C) 2005 Christophe Choquet (cchoquet@users.sourceforge.net)   

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_string.hh"
#include "as_array.hh"

typedef void (ASString::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASString::init,			// 0
	&ASString::charAt,			// 1
	&ASString::charCodeAt,		// 2
	&ASString::concat,			// 3
	&ASString::fromCharCode,	// 4
	&ASString::indexOf,			// 5
	&ASString::lastIndexOf,		// 6
	&ASString::slice,			// 7
	&ASString::split,			// 8
	&ASString::substr,			// 9
	&ASString::substring,		// 10
	&ASString::toLowerCase,		// 11
	&ASString::toUpperCase,		// 12
	// Need to be at the end.
	&ASString::addProperty,		
	&ASString::registerClass,	
	&ASString::AStoString,		
	&ASString::unwatch,			
	&ASString::valueOf,			
	&ASString::watch
};

/**
 * The constructors
 **/
ASString::ASString(Script *s, Graphics *g) : ASObject(s,g,L"String") {
}
ASString::ASString(Script *s, Graphics *g, wstring value) : ASObject(s,g,L"String") {
	this->value = value;
	this->setMemberNoClone(L"length",new IntegerObject(value.size()));
}

/**
 * The destructor
 **/
ASString::~ASString() {
}

/**
 * Return a function descriptor
 */
Function *ASString::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"String") 	{func->setFunctionIndex(0); func->setNumParams(1); }
	if (memberName == L"charAt") 	{func->setFunctionIndex(1); func->setNumParams(1); }
	if (memberName == L"charCodeAt"){func->setFunctionIndex(2); func->setNumParams(1); }
	if (memberName == L"concat") 	{func->setFunctionIndex(3); func->setNumParams(10); } // unlimited
	if (memberName == L"fromCharCode"){func->setFunctionIndex(4); func->setNumParams(10); } // unlimitted
	if (memberName == L"indexOf") 	{func->setFunctionIndex(5); func->setNumParams(2); }
	if (memberName == L"lastIndexOf"){func->setFunctionIndex(6); func->setNumParams(2); }
	if (memberName == L"slice") 	{func->setFunctionIndex(7); func->setNumParams(2); }
	if (memberName == L"split") 	{func->setFunctionIndex(8); func->setNumParams(2); }
	if (memberName == L"substr") 	{func->setFunctionIndex(9); func->setNumParams(2); }
	if (memberName == L"substring")	{func->setFunctionIndex(10); func->setNumParams(2); }
	if (memberName == L"toLowerCase"){func->setFunctionIndex(11); func->setNumParams(0); }
	if (memberName == L"toUpperCase"){func->setFunctionIndex(12); func->setNumParams(0); }
	func = finishFunctionSetup(memberName,13,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASString::callFunction(Function *func) {
	this->objectInstance = func->getObjectInstance();
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 * String constructor.
 **/
void ASString::init() {
	value = L"";
	if (numPassedParams > 0) {
		value = script->getRegister(1)->toString();
	}	
	this->setMemberNoClone(L"length",new IntegerObject(value.size()));
}

/**
 *  An integer that specifies the position of a character in the string. The first character is index indicated by 0, 
 * and the last character is indicated by my_str.length-1.
 **/
void ASString::charAt() {
	wstring res = L"";
	if (numPassedParams > 0) {
		int index = int(script->getRegister(1)->toNumeric());
		if (index >= 0 && index < int(value.size())) res = value.substr(index,1);
	}
	script->push(new StringObject(res));		
}


/**
 * Returns the char code
 **/
void ASString::charCodeAt() {
	int res = -1;
	if (numPassedParams > 0) {
		int index = int(script->getRegister(1)->toNumeric());
		if (index >= 0 && index < (int)value.size()) res = index;
	}
	if (res != -1) script->push(new IntegerObject(value.at(res)));	
	else script->push(new StringObject(L"NaN"));
}


/**
 * Zero or more values to be concatenated.
 **/
void ASString::concat() {
	wstring res = value;
	for (int i=0; i< numPassedParams; i++) {
		res += script->getRegister(i+1)->toString();
	}
	script->push(new StringObject(res));
}


/**
 * returns a string made up of the characters represented by the ASCII values in the parameters.
 **/
void ASString::fromCharCode() {
	wstring res = L"";
	for (int i=0; i< numPassedParams; i++) {
		res += char(script->getRegister(i+1)->toNumeric());
	}
	script->push(new StringObject(res));
}


/**
 * Searches the string and returns the position of the first occurrence of substring found at or after startIndex within the calling string. 
 * If substring is not found, the method returns -1.
 **/
void ASString::indexOf() {
	if (numPassedParams == 0) { script->push(new StringObject(L"undefined")); return ; }
	wstring s2 = script->getRegister(1)->toString();
	int index = 0;
	if (numPassedParams > 1) index = int(script->getRegister(2)->toNumeric());
	wstring::size_type pos = value.find(s2,index);
	if (pos != wstring::npos) script->push(new IntegerObject(pos)); else script->push(new IntegerObject(-1));
}


/**
 * Same as indexOf but from the end of the string
 **/
void ASString::lastIndexOf() {
	if (numPassedParams == 0) { script->push(new StringObject(L"undefined")); return ; }
	wstring s2 = script->getRegister(1)->toString();
	int index = value.size();
	if (numPassedParams > 1) index = int(script->getRegister(2)->toNumeric());
	wstring::size_type pos = value.find_last_of(s2,index);
	if (pos != wstring::npos) script->push(new IntegerObject(pos)); else script->push(new IntegerObject(-1));
}


/**
 * returns a string that includes the start character and all characters up to (but not including) the end character. 
 * The original String object is not modified. If the end parameter is not specified, the end of the substring is the end of the string. 
 * If the value of start is greater than or equal to the value of end, the method returns an empty string.
 **/
void ASString::slice() {
	if (numPassedParams == 0) { script->push(new StringObject(L"undefined")); return ; }
	if (numPassedParams == 1) {
		int index = int(script->getRegister(1)->toNumeric());
		if (index < 0) index = value.size() + index;
		if (index < (int)value.size()) script->push(new StringObject(value.substr(index))); else script->push(new StringObject(L""));
	} else {
		int index = int(script->getRegister(1)->toNumeric());
		if (index < 0) index = value.size() + index;
		int end = int(script->getRegister(2)->toNumeric());
		if (end > (int)value.size()) end = value.size();
		if (index < (int)value.size()) script->push(new StringObject(value.substr(index,end))); else script->push(new StringObject(L""));
	}
}


/**
 * splits a String object into substrings by breaking it wherever the specified delimiter parameter occurs, and returns the substrings in an array. 
 * If you use an empty string ("") as a delimiter, each character in the string is placed as an element in the array, as in the following code.
 **/
void ASString::split() {
	if (numPassedParams <= 0) { script->push(new StringObject(value)); return ; }
	wstring separator = script->getRegister(1)->toString();
	int limit = value.size();
	if (numPassedParams > 1) limit = int(script->getRegister(2)->toNumeric());
	ASArray *array = new ASArray(script,graphics);
	bool flush = false;
	wstring tmp = L"";
	int j=0;
	for (wstring::size_type i=0; i< value.size() && j < limit; i++) {
		if (separator == L"") { 
			tmp += value.at(i); flush = true; 
		} else {
			if (separator.at(0) != value.at(i)) {
				tmp += value.at(i);
			} else {
				flush = true; 
			}
		}
		if (flush) {
			char x[10]; sprintf(x,"%d",j++);
			array->setMember(stringTowstring(x), new StringObject(tmp));
			tmp = L"";
			flush = false;
		}		
	}
	if (tmp.size() > 0) {
		char x[10]; sprintf(x,"%d",j++);
		array->setMember(stringTowstring(x), new StringObject(tmp));
	}
	script->push(array);
}



/**
 * substr(start, [length])
 **/
void ASString::substr(){
	if (numPassedParams == 0) { script->push(new StringObject(L"undefined")); return ; }
	wstring s=L"";
	int start = int(script->getRegister(1)->toNumeric());
	if (start < 0) start = value.size() + start;
	if (numPassedParams == 1) {
		if (start < (int)value.size()) s = value.substr(start);
	} else {
		int length = int(script->getRegister(2)->toNumeric());
		if (start < (int)value.size()) {
			if (start+length >= (int)value.size()) length = value.size();
			s = value.substr(start,length);
		}
	}
	script->push(new StringObject(s));	
};

/**
 * substring(start, [end])
 **/
void ASString::substring() {
	if (numPassedParams == 0) { script->push(new StringObject(L"undefined")); return ; }
	wstring s=L"";
	int start = int(script->getRegister(1)->toNumeric());
	if (start < 0) start = 0;
	if (numPassedParams == 1) {
		if (start < (int)value.size()) s = value.substr(start);
	} else {
		int end = int(script->getRegister(2)->toNumeric());
		if (end<0) { end = start; start = 0; }
		if (start > end) end=start; 
		if (start >= 0 && start < (int)value.size()) {
			if (end >= (int)value.size()) end = value.size();
			s = value.substr(start,end-start);
		}
	}
	script->push(new StringObject(s));	
}


/**
 * returns a copy of the String object, with all of the uppercase characters converted to lowercase. The original value is unchanged.
 **/
void ASString::toLowerCase() {
	wstring res =L"";
	for (wstring::size_type i=0; i< value.size(); i++) res += tolower(value.at(i));
	script->push(new StringObject(res));	
}


/**
 * returns a copy of the String object, with all of the lowercase characters converted to uppercase. The original value is unchanged.
 **/
void ASString::toUpperCase() {
	wstring res =L"";
	for (wstring::size_type i=0; i< value.size(); i++) res += toupper(value.at(i));
	script->push(new StringObject(res));	
}

/**
 * Returns the string as ... as string
 **/
void ASString::AStoString() {
	script->push(new StringObject(value));
}

/**
 * Returns the string as ... as string
 **/
wstring ASString::toString() const{
	return value;
}

/**
 * Return the string value property
 **/
ValueObject*ASString::getMainMember() {
	return new StringObject(value);
}
