/**
Macromedia Flash BaseRenderer interface implementation.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "brenderer.hh" 

BaseRenderer::BaseRenderer(BaseTexture *texture) {
	this->texture = texture;
	backGroundColor.red = backGroundColor.green = backGroundColor.blue = 0;
	newTexCoords = NULL;
}

BaseRenderer::~BaseRenderer() {
	if (newTexCoords!= NULL) free(newTexCoords);
}

/**
 * Apply a transformation matrix.
 * A transformation matrix is given by:
 * x' = x * ScaleX + y * RotateSkew1 + TranslateX
 * y' = x * RotateSkew0 + y * ScaleY + TranslateY
 *
 * We create here a transform matrix, and multiply it on the stack
 **/
void BaseRenderer::applyMatrixMaths(float *mat, MATRIX *m) {
	lineRatio = 1.25;
	for (int i=0; i< 4; i++) 
		for (int j=0; j <4; j++)
			mat[i*4+j] = i==j  ? 1.0 : 0.0;

	mat[3*4+0] = m->translateX;
	mat[3*4+1] = m->translateY;
	if (m->hasScale ) {
		mat[0*4+0] = (float)m->scaleX/(float)0x10000; 
		mat[1*4+1] = (float)m->scaleY/(float)0x10000;
		if (m->scaleX != 0 || m->scaleY != 0) {
			// This width is false, but it is difficult (and expensive) to have the correct one. How ?
			lineRatio = sqrt(mat[0*4+0]*mat[0*4+0] + mat[1*4+1]*mat[1*4+1]);
			if (lineRatio < 1.25) lineRatio = 1.25;
		}
	}
	if (m->hasRotate) {
		mat[0*4+1] = (float)m->rotateSkew0/(float)0x10000; 
		mat[1*4+0] = (float)m->rotateSkew1/(float)0x10000;
	}
	//printf("%f %f   %f %f    %f %f \n",mat[3*4+0],mat[3*4+1],mat[0*4+0],mat[1*4+1],mat[0*4+1],mat[1*4+0]);
}

/**
 * If we don't have Vertex Shader, we need to recompute the colors given the CXFORMWITHALPHA
 ***/
void BaseRenderer::updateTextureNoShader(DrawingSequence *d, DrawingEntry *de, bool isTexture,int texDim) {
	int coef = 3;
	if (de->drawingType == DrawingLines) coef = 2;
	if (de->drawingType == DrawingFontChars) coef = 6;
	for (int i=0; i< coef*de->nbElements && i < d->getNbPoints(); i++) {
		RGBA originalColor;
		float tx,ty;
		if (isTexture) {
			originalColor = texture->getColorByTextureCoords(de->texNum,de->getTexCoords()[i*texDim],texDim == 2 ? de->getTexCoords()[i*2+1] : 0);
		} else {
			int colorIndex = int (de->getTexCoords()[i*2]*256.0 + de->getTexCoords()[i*2+1]*65536.0);
			originalColor = texture->getColorByIndex(colorIndex);
		}
		RGBA newColor;
		newColor.red   = long(originalColor.red *cxAlpha[0]   + secondPassColor[0]*255.0);
		newColor.green = long(originalColor.green *cxAlpha[1] + secondPassColor[1]*255.0);
		newColor.blue  = long(originalColor.blue *cxAlpha[2]  + secondPassColor[2]*255.0);
		newColor.alpha = long(originalColor.alpha *cxAlpha[3] + secondPassColor[3]*255.0);
		texture->registerColor(&newColor);
		int loc = texture->getColorTextureIndex(&newColor);
		tx = (loc%256)/256.0f;
		ty = float(loc/256)/256.0f;
		if (texDim == 2) {
			newTexCoords[i*2]   = tx;
			newTexCoords[i*2+1] = ty;
		} else {
			newTexCoords[i]   = tx;
		}
	}
}

/**
 * Implements CXFORMWITHALPHA rendering.
 **/
void BaseRenderer::applyCxFormWithAlpha(CXFORMWITHALPHA *c ) {
	if (c->hasAddTerms) { 
		float mat[4];
		mat[0] = float(c->redAddTerm) /256.0; 
		mat[1] = float(c->greenAddTerm) /256.0; 
		mat[2] = float(c->blueAddTerm) /256.0; 
		mat[3] = float(c->alphaAddTerm) /256.0; 
		if (mat[0] != 0 || mat[1] != 0 || mat[2] != 0 || mat[3] != 0) {
			memcpy(secondPassColor,mat,sizeof(mat));
			hasCxAlpha = true;
			//printf("Second pass %f %f %f %f\n", secondPassColor[0],secondPassColor[1],secondPassColor[2],secondPassColor[3]);
		}
	} else memset(secondPassColor,0,sizeof(secondPassColor));

	if (c->hasMultTerms) {
		cxAlpha[0] = float(c->redMultTerm) /256.0; 
		cxAlpha[1] = float(c->greenMultTerm) /256.0;
		cxAlpha[2] = float(c->blueMultTerm) /256.0; 
		cxAlpha[3] = float(c->alphaMultTerm) /256.0; 
		//printf("CxFormWithAlpha %f %f %f %f\n",cxAlpha[0],cxAlpha[1],cxAlpha[2],cxAlpha[3]);
		hasCxAlpha = true;
	} else cxAlpha[0]=cxAlpha[1]=cxAlpha[2]=cxAlpha[3]=1;
}

/**
 * Disable CxAlpha
 **/
void BaseRenderer::noCxFormWithAlpha() {
	hasCxAlpha = false;
	cxAlpha[0]=cxAlpha[1]=cxAlpha[2]=cxAlpha[3]=1.0;
	secondPassColor[0]=secondPassColor[1]=secondPassColor[2]=secondPassColor[3]=0.0;
}

/**
 * Push a new CxFormWithAlpha
 **/
void BaseRenderer::pushCxFormWithAlpha(CXFORMWITHALPHA* c) {
	if (cxFormWithAlphaStack.size() == 0) cxFormWithAlphaStack.push(*c);
	else {
		CXFORMWITHALPHA d = cxFormWithAlphaStack.top();
		cxFormWithAlphaStack.push(multiplyCxFormWithAlpha(c,&d));
	}
	applyCxFormWithAlpha(&cxFormWithAlphaStack.top());
}

/**
 * Restore original CxFormWithAlpha
 **/
void BaseRenderer::popCxFormWithAlpha() {
	cxFormWithAlphaStack.pop();
	if (cxFormWithAlphaStack.size() > 0) applyCxFormWithAlpha(&cxFormWithAlphaStack.top()); else noCxFormWithAlpha();
}

void BaseRenderer::noMatrix() {
	lineRatio = 1; 
}

void BaseRenderer::setBackGroundColor(RGB color) { 
	backGroundColor = color;
};
