/**
Macromedia Flash  - Event Manager implementation.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "event_manager.hh"
#include "../ASLib/as_text_field.hh"
#include "../ASLib/as_linker.hh"
#define XK_MISCELLANY
#include <X11/keysymdef.h>

extern bool debug;

/**
 * The constructor
 **/
EventManager::EventManager(ObjectStore *p, Graphics *g, SwfEngine *e) {
	this->objectStore = p;
	this->graphics = g;
	this->engine = e;
}

EventManager::~EventManager() {
	///interactions.clear();
	for (list<CallBackRequest*>::const_iterator i = timerCallbacks.begin(); i!=timerCallbacks.end(); ++i) delete *i;
	timerCallbacks.clear();
}

/**
 * Register a new clip Action
 **/
void EventManager::registerClipAction(ScriptObject *sc,ClipActions *c){
	//wcout << L"REGISTER clip action for " << *sc << L"\n";
	EVENT_MANAGER_ENTRY e;
	e.clipActions = c;
	e.so = sc;
	events[buildKey(sc)] = e;
}

/**
 * Process an onLoadEvent
 **/
void EventManager::doOnLoad(ScriptObject *sc) {
	EVENT_MANAGER_ENTRY e = events[buildKey(sc)];
	ClipActions *c = e.clipActions;
	if (c == NULL) {
		// Check is we have an onLoad member
		ValueObject *v = sc->getMember(L"onLoad");
		if (v != NULL && v->getThisObjectInstanceName() == L"function") {
			Function *f = (Function*)v;
			executeFunction(sc,f);
		}
	} else {
		if (c->getAllEventFlags().clipEventLoad || c->getAllEventFlags().clipEventInitialize) {
			for (list<ClipActionRecord*>::iterator it=c->getClipActionRecords()->begin(); it != c->getClipActionRecords()->end(); it++) {
				ClipActionRecord *cre = *it;
				if (cre->getClipEventFlags().clipEventLoad || cre->getClipEventFlags().clipEventInitialize) {
					if (debug ) wcout << L"EXECUTE ON LOAD " << e.so/*->getMovieName()*/ << L"\n";
					executeClipActionScript(e.so,cre,true);
					((ASObject*)e.so)->setLoaded(true);
				}
			}
		}
	}
}

/**
 * Process an ALL onLoadEvent
 *
 * Events can be defined two ways:
 * - an actionClip object may be associated to an object,
 * - the event can be defined as a member by action script
 **/
void EventManager::doOnLoad() {
	for (StringEventManagerEntryHashMap::const_iterator i = events.begin(); i != events.end(); ++i) {
		EVENT_MANAGER_ENTRY e = i->second;
		ClipActions *c = e.clipActions;
		if (c == NULL) return;
		if (!((ASObject*)e.so)->getIsLoaded()) {
			if (c->getAllEventFlags().clipEventLoad || c->getAllEventFlags().clipEventInitialize) {
				//engine->getScript()->prepareExecEnv(e.so);
				for (list<ClipActionRecord*>::iterator it=c->getClipActionRecords()->begin(); it != c->getClipActionRecords()->end(); it++) {
					ClipActionRecord *cre = *it;
					if (cre->getClipEventFlags().clipEventLoad || c->getAllEventFlags().clipEventInitialize) {
						if (debug ) wcout << L"EXECUTE ON LOAD " << e.so->getMovieName();
						executeClipActionScript(e.so,cre,true);
					}
				}
				//engine->getScript()->leaveExecEnv();
			}
			((ASObject*)e.so)->setLoaded(true);
		}
	}
}

/**
 * Process an onLoadEnterFrame
 **/
void EventManager::doOnEnterFrame(ScriptObject *sc) {
	// First look if we have an action script declared for this object
	EVENT_MANAGER_ENTRY e = events[buildKey(sc)];
	ClipActions *c = e.clipActions;
	if (c != NULL) {
		if (c->getAllEventFlags().clipEventEnterFrame) {
			engine->getScript()->prepareExecEnv(e.so);
			for (list<ClipActionRecord*>::iterator it=c->getClipActionRecords()->begin(); it != c->getClipActionRecords()->end(); it++) {
				ClipActionRecord *cre = *it;
				if (cre->getClipEventFlags().clipEventEnterFrame) {
					//if (debug) wcout << L"ON ENTERFRAME: "<< *sc << L" "<< ((ASObject*)sc)->getMovieName() << L"(" << sc << L")" << L"\n";
					executeClipActionScript(sc,cre,true);
				}
			}
			engine->getScript()->leaveExecEnv();
		}
	}

	//wcout << L"doOnEnterFrame " << sc << L" " << sc->getMovieName() << L"\n";
	// Then look if we have a member called "onEnterFrame"
	ValueObject *v = sc->getMember(L"onEnterFrame");
	if (v != NULL) {
		//if (debug) wcout << "\nProcess onEnterFrame on " << *sc << L" (" << sc->getMovieName() << L")\n";
		Function *func = (Function*)v;
		engine->getScript()->prepareExecEnv(sc);
		objectStore->setPoolActionContext(func->getConstantPoolKey());
		if (func->getIsInlined()) objectStore->setCurrentTargetPath(func->getTargetPath());
		engine->executeCode(func->getCode(), func->getCodeSize());
		engine->getScript()->leaveExecEnv();
	}
}

/**
 * Process an doOnMouseDown event type
 **/
void EventManager::doOnMouseDown(ScriptObject *sc) {
	EVENT_MANAGER_ENTRY e = events[buildKey(sc)];
	ClipActions *c = e.clipActions;
	if (c != NULL && c->getAllEventFlags().clipEventMouseDown) {
		engine->getScript()->prepareExecEnv(e.so);
		for (list<ClipActionRecord*>::iterator it=c->getClipActionRecords()->begin(); it != c->getClipActionRecords()->end(); it++) {
			ClipActionRecord *cre = *it;
			if (cre->getClipEventFlags().clipEventMouseDown) {
				if (debug ) wcout << L"EXECUTE ON MOUSE DOWN ";
				executeClipActionScript(e.so,cre,false);
			}
		}
		engine->getScript()->leaveExecEnv();
	}
}

/**
 * Execute a clip action record script
 **/
void EventManager::executeClipActionScript(ScriptObject *sc,ClipActionRecord *cr, bool hasLocalScope) {
	engine->getScript()->prepareExecEnv(sc);
	objectStore->setHasLocalVariableScope(hasLocalScope);
	engine->executeCode(cr->getCode());
	objectStore->setHasLocalVariableScope(true);
	engine->getScript()->leaveExecEnv();
}

/*
 * Execute a function as an event
 **/
void EventManager::executeFunction(ScriptObject *sc,Function *f){
	engine->getScript()->prepareExecEnv(sc,f->getConstantPoolKey());
	engine->executeCode(f->getCode(), f->getCodeSize());
	engine->getScript()->leaveExecEnv();
}

/**
 * Update object object bouding box for user input
 * 
 * We perform some simplifications here as we suppose that the transformed box is still a rectangle.
 * The user selection area will be altered if there are some rotation to the object.
 **/
void EventManager::updateSelectionArea(int depth,ASInterface *a, int charId) {
	//wcout << L"EventManager::updateSelectionArea " << depth << L" " << *a << L"\n"; 
	if (a->getClassName() == L"Button" || a->getClassName() == L"TextField" || a->getClassName() == L"Sprite") {
		bool hasInteraction = true;
		if (a->getClassName() == L"TextField") {
			ASTextField *m = (ASTextField*) a;
			hasInteraction = !m->isStatic();
		}
		if (hasInteraction) {
			float transform[4][4];
			engine->getOglRenderer()->getTransformMatrix((float*)transform);
			a->setInverseMatrix((float*)transform, charId);

			// Register Target path & Object
			INTERACTION inter;
			inter.destObject = a;
			inter.destObjectTargetPath = objectStore->getCurrentTargetPath();
			interactions[depth] = inter;
		}
	} else {
		// Register inverse matrix in other cases. We need it to be able to manage _mousex and _mousey
		// This matrix is used to convert sceen mouse coordinates to onject coordinates.
		float transform[4][4];
		engine->getOglRenderer()->getTransformMatrix((float*)transform);
		a->setInverseMatrix((float*)transform,charId);
	}
}

/**
 * Mouse handler for click event
 **/
bool EventManager::mouseClickCallback(int x, int y,int mouseAction) {
	mouseX = x;
	mouseY = y;
	bool needRedraw = false;
	EventDataCallback eventDataCallback;
	eventDataCallback.x = x;
	eventDataCallback.y = y;
	eventDataCallback.mouseAction = mouseAction;

	// The interaction may be cleared in case of a goto	intruction. We then need to exit, since the iterator is wrong.
	quitRequired = false;
	for (IntINTERACTIONHashMap::iterator it = interactions.begin(); !quitRequired && it !=interactions.end(); it++) {
		INTERACTION inter = (*it).second;
		ASInterface *a = inter.destObject;
		if (a == NULL) continue;

		// Set the target path of the selected object.
		wstring oldTargetPath = objectStore->getCurrentTargetPath();
		objectStore->setCurrentTargetPath(inter.destObjectTargetPath);

		// Check if inside bounding box, and call the handler
		if (a->isInSelectionBox(x,y)) {
			//printf("TRANSFORMED: %d %d %s\n",x,y,a->getClassName().c_str());
			if (mouseAction == MouseActionPress)  {
				needRedraw|=a->callInternalEvent("onPress",&eventDataCallback);
				doOnMouseDown(a);
			}
			if (mouseAction == MouseActionRelease) needRedraw|=a->callInternalEvent("onRelease",&eventDataCallback); 
			if (mouseAction == MouseActionOver)    needRedraw|=a->callInternalEvent("onRollOver",&eventDataCallback);
		} else {
			needRedraw|=a->callInternalEvent("onRollOut",NULL);
		}			
		objectStore->setCurrentTargetPath(oldTargetPath);
	}
	return needRedraw;
}

/**
 * Event handler for keyboard input
 *
 **/
void EventManager::keyboardCallback(int key, char *code) {
	key = translateKeys(key);
	lastKeyPressed = key;
	iskeyPressed = true;
printf("Keyboard, key:%d\n",key);

	// Look for buttons
	for (IntINTERACTIONHashMap::iterator it = interactions.begin(); it !=interactions.end(); it++) {
		INTERACTION inter = (*it).second;
		ASInterface *a = inter.destObject;
		if (a == NULL) continue;
		if (a->handleThisKey(key)) {
			puts("Handle this key");
		}
	}

	// Look for actions - those are defined as ClipActions
	for (StringEventManagerEntryHashMap::const_iterator i = events.begin(); i != events.end(); ++i) {
		EVENT_MANAGER_ENTRY e = i->second;
		ClipActions *c = e.clipActions;
		if (c == NULL) continue;
		ClipEventFlags cef = c->getAllEventFlags();
		if (cef.clipEventKeyDown) {
			for (list<ClipActionRecord*>::const_iterator j = c->getClipActionRecords()->begin(); j != c->getClipActionRecords()->end(); ++j) {
				ClipActionRecord *car = *j;
				ClipEventFlags cefj = car->getClipEventFlags();
				if (cef.clipEventKeyDown) {
					executeClipActionScript(e.so,car,false);
				}
			}
		}
	}

	// Look for Key.onKeyDown event support.
	ASObject *f = (ASObject*)engine->getScript()->getASLinker()->getClassSingleton(L"Key");
	ASObject *listener = (ASObject *) f->getListener(L"onKeyDown");
	if (listener != NULL) {
		// The listener object is here, look for the event member
		Function *func = (Function*)  listener->getMember(L"onKeyDown");
		if (func != NULL) {
			executeFunction(listener->getParent(),func);
		}
	}
}

/**
 * Translate keyCode.
 **/
int	EventManager::translateKeys(int keyCode) {
	int key = keyCode;
	switch (keyCode) {
		case XK_BackSpace:
			key = 8;
			break;
		case XK_Caps_Lock:
			key = 20;
			break;
		case XK_Control_L:
		case XK_Control_R:
			key = 17;
			break;
		case XK_Delete:
			key = 46;
			break;
		case XK_Down:
			key = 40;
			break;
		case XK_End:
			key = 35;
			break;
		case XK_KP_Enter:
		case XK_Return:
			key = 13;
			break;
		case XK_Escape:
			key = 27;
			break;
		case XK_Home:
			key = 36;
			break;puts("TIMER CALL REQUESTED");

		case XK_Insert:
			key = 45;
			break;
		case XK_Left:
			key = 37;
			break;
		case XK_Page_Down:
			key = 34;
			break;
		case XK_Page_Up:
			key = 33;
			break;
		case XK_Right:
			key = 39;
			break;
		case XK_Shift_L:
		case XK_Shift_R:
			key = 16;
			break;
		case XK_Up:
			key = 38;
			break;
	}
	return key;
}

/**
 * Clone actions
 **/
void EventManager::cloneClipActions(ScriptObject*s1,ScriptObject*s2){
	if (events[buildKey(s1)].clipActions != NULL) {
		events[buildKey(s2)] = events[buildKey(s1)];
	}
}

/**
 * Add a timer callback entry
 **/
void EventManager::addTimerCallbackEntry(CallBackRequest *c) {
	for (list<CallBackRequest*>::const_iterator i = timerCallbacks.begin(); i!=timerCallbacks.end(); ++i) {
		if (*i == c) return;
	}
	timerCallbacks.push_back(c);
}

/**
 * Check timer clallback
 **/
void EventManager::doTimerCallbacks() {
	for (list<CallBackRequest*>::const_iterator i = timerCallbacks.begin(); i!=timerCallbacks.end(); ++i) {
		CallBackRequest *c = *i;
		if (c->requestCall()) {
			//puts("TIMER CALL REQUESTED");
			list<ValueObject*>*params = c->getParams();
			for (list<ValueObject*>::const_iterator i = params->end(); i != params->begin(); --i) {
				engine->getScript()->push((*i)->clone());
			}
			engine->getScript()->prepareExecEnv(engine->getGlobal());
			engine->getScript()->push(new IntegerObject(c->getParams()->size()));
			engine->getScript()->push(new StringObject(c->getFunctionName()));

			int k = 0;
			UI8 *code = (UI8*)&k;
			engine->getScript()->doActionCallFunction(&code);
			engine->executeCode(code);
			engine->getScript()->leaveExecEnv();
			//puts("/TIMER CALL REQUESTED");
		}
	}
}

/**
 * Register a new CallBack request
 **/
CallBackRequest::CallBackRequest(ObjectStore*o,wstring fn, int interval, list<ValueObject*> params) {
	this->objectStore = o;
	this->functionName = fn;
	this->interval = interval;
	this->params = params;
	time = getTime();
}

/**
 * destructor
 **/
CallBackRequest::~CallBackRequest() {
	for (list<ValueObject*>::const_iterator i=params.begin(); i!= params.end(); ++i) objectStore->removeObject(*i);
	params.clear();
}

/**
 * Return current time in milliseconds
 **/
long CallBackRequest::getTime() {
	struct timeval tEnd;
	gettimeofday(&tEnd,NULL);
	uint32_t t;
	t = tEnd.tv_usec/1000;
	t += tEnd.tv_sec*1000;
	return t;
}

/**
 * If the calling time has occurs return true and update next event time
 **/
bool CallBackRequest::requestCall() {
	long t = getTime();
	if (t > time+interval) {
		time += interval;
		return true;
	} else {
		return false;
	}
}

