/**
Macromedia Flash File Uncompress lib.
Copyright (C) 2004  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "uncompress.hh"
#include <pthread.h>

/**
 * The jpeg library use C style call handlers taht updates static variables.
 * That's not thread safe, so we need semaphore here to prevent two simultaneous uncompress to occur at the same time.
 **/
static pthread_mutex_t  synchro;
static unsigned char *inputData;		// Thread protected by synchro semaphore

static void errorExit(j_common_ptr info) {
	(*info->err->output_message) (info);
	longjmp(((MyErrorHandler *)info->err)->setjmp_buffer, 1);
}

// Methods for Source data manager
static void initSource(struct jpeg_decompress_struct *cInfo) {
	cInfo->src->bytes_in_buffer = 0;
}

static boolean fillInputBuffer(struct jpeg_decompress_struct *cInfo) {
	cInfo->src->next_input_byte = inputData;
	cInfo->src->bytes_in_buffer = 1;
	inputData++;
	return true;
}

static void skipInputData(struct jpeg_decompress_struct *cInfo, long count) {
	cInfo->src->bytes_in_buffer = 0;
	inputData += count;
}

static boolean resyncToRestart(struct jpeg_decompress_struct *cInfo, int desired) {
	return jpeg_resync_to_restart(cInfo, desired);
}

static void termSource(struct jpeg_decompress_struct *cInfo) {
}

/**
 * Uncompress a JPEG stream
 **/
bool UncompressTools::uncompressJPEGData(UI8 *stream,Image *img, bool read_alpha, long offset) {
	struct jpeg_decompress_struct cInfo;
	struct jpeg_source_mgr mySrcMgr;
	MyErrorHandler errorMgr;
	JSAMPROW buffer[1];
	unsigned char *ptrPix;
	int stride;
	
	// Kludge to correct some corrupted files
	if (stream[1] == 0xd9 && stream[3] == 0xd8) {
		stream[3] = 0xd9;
		stream[1] = 0xd8;
	}
	pthread_mutex_lock(&synchro);

	// Setup error handler
	cInfo.err = jpeg_std_error(&errorMgr.pub);
	errorMgr.pub.error_exit = errorExit;

	if (setjmp(errorMgr.setjmp_buffer)) {
		jpeg_destroy_decompress(&cInfo);
		return false;
	}

	inputData = stream;
	jpeg_create_decompress(&cInfo);

	// Setup source manager structure
	mySrcMgr.init_source = initSource;
	mySrcMgr.fill_input_buffer = fillInputBuffer;
	mySrcMgr.skip_input_data = skipInputData;
	mySrcMgr.resync_to_restart = resyncToRestart;
	mySrcMgr.term_source = termSource;

	// Set default source manager
	cInfo.src = &mySrcMgr;

	jpeg_read_header(&cInfo, FALSE);

	jpeg_read_header(&cInfo, TRUE);
	cInfo.quantize_colors = TRUE;	// Create colormapped image
	jpeg_start_decompress(&cInfo);

	// Set objet dimensions
	img->setHeight(cInfo.output_height);
	img->setWidth(cInfo.output_width);
    img->setBpl(img->getWidth());
	UI8 *pixels = new UI8[img->getWidth()*img->getHeight()];
	if (pixels == NULL) {
		jpeg_finish_decompress(&cInfo);
		jpeg_destroy_decompress(&cInfo);
		return false;
	}
	ptrPix = pixels;

	stride = cInfo.output_width * cInfo.output_components;
	buffer[0] = (JSAMPROW)malloc(stride);

	while (cInfo.output_scanline < cInfo.output_height) {
		jpeg_read_scanlines(&cInfo, buffer, 1);
		memcpy(ptrPix,buffer[0],stride);
		ptrPix+= stride;
	}
	free(buffer[0]);

	RGBA *colormap = new RGBA[cInfo.actual_number_of_colors];
	if (colormap == NULL) {
		jpeg_finish_decompress(&cInfo);
		jpeg_destroy_decompress(&cInfo);
		return false;
	}
	int nbColors = cInfo.actual_number_of_colors;

	// There is something really really strange here. 
	// Sometimes the range is set to 0x00-0x7f
	// Scale it if necessary. Need to understand what's happening!!!!!
	int colorMax = 0;
	if (cInfo.out_color_space == JCS_RGB) {
		for(int n=0; n < nbColors; n++) {
			colormap[n].red = cInfo.colormap[0][n];
			colormap[n].green = cInfo.colormap[1][n];
			colormap[n].blue = cInfo.colormap[2][n];
			colormap[n].alpha = 255;
			colorMax = colorMax < colormap[n].red ? colormap[n].red : colorMax;
			colorMax = colorMax < colormap[n].green ? colormap[n].green : colorMax;
			colorMax = colorMax < colormap[n].blue ? colormap[n].blue : colorMax;
		}
		if (colorMax < 0x7f) {
			for(int n=0; n < nbColors; n++) {
				colormap[n].red += 0x7f;
				colormap[n].green += 0x7f;
				colormap[n].blue += 0x7f;
			}
		}
	} else {
		puts("TODO: OTHER JPEG COLOR FORMAT");
	}

	jpeg_finish_decompress(&cInfo);
	jpeg_destroy_decompress(&cInfo);

	if (read_alpha) {
		if (!buildJpegAlpha(img,  stream + offset)) {
			return false;
		}
	}

	pthread_mutex_unlock(&synchro);
	img->setNbColors(nbColors);
	img->setPixels(pixels);
	img->setColormap(colormap);
	return true;
}

/**
 * Uncompress the alpha buffer
 **/
bool UncompressTools::buildJpegAlpha(Image *img, UI8 *buffer) {
	z_stream	stream;
	int		status;
	UI8		*data;

	data = new UI8[img->getWidth()*img->getHeight()];
	if (data == NULL) return false;

    stream.next_in = buffer;
    stream.avail_in = 1;
    stream.next_out = data;
    stream.avail_out = img->getWidth()*img->getHeight();
    stream.zalloc = Z_NULL;
    stream.zfree = Z_NULL;
            
    status = inflateInit(&stream);
    while (1) {
        status = inflate(&stream, Z_SYNC_FLUSH) ;
        if (status == Z_STREAM_END) {
            break;
        }
        if (status != Z_OK) {
            printf("Zlib data error : %s\n", stream.msg);
			delete []data;
            return false;
        }
        stream.avail_in = 1;
    }

    inflateEnd(&stream);         
    img->setAlphaBuf(data);
    return true;
}

/**
 * Uncompress bitslossless data
 **/
bool UncompressTools::buildFromZlibData(UI8 *buffer,Image *img, int width, int height, int format, int tableSize, int tableHasAlpha) {
	z_stream	stream;
	int		status;
	unsigned char  *data;
	int		elementSize;

	img->setWidth(width);
	img->setHeight(height);
	img->setBpl(width);

	if (tableHasAlpha) {
		elementSize = 4;	// Cmap is RGBA
	} else {
		elementSize = 3;	// Cmap is RGB
	}

	stream.next_in = buffer;
	stream.avail_in = 1;
	stream.zalloc = Z_NULL;
	stream.zfree = Z_NULL;

	tableSize++;
	RGBA *colormap = NULL;

	// Uncompress Color Table
	if (format == 3) {
		long n;

		// Ajust width for 32 bit padding
		width = (width+3)/4*4;
		img->setWidth(width);
		img->setBpl(width);
		img->setDepth(1);
		UI8 *colorTable = new UI8[tableSize*elementSize];
		if (colorTable == NULL) return false;

		stream.next_out = colorTable;
		stream.avail_out = tableSize*elementSize;
		inflateInit(&stream);

		while (1) {
			status = inflate(&stream, Z_SYNC_FLUSH);
			if (status == Z_STREAM_END) {
					break;
			}
			if (status != Z_OK) {
				printf("Zlib cmap error : %s\n", stream.msg);
				return false;
			}
			stream.avail_in = 1;
			// Colormap if full
			if (stream.avail_out == 0) {
				break;
			}
		}

		img->setNbColors(tableSize);

		colormap = new RGBA[img->getNbColors()];
		if (colormap == NULL) {
			delete []colorTable;
			return false;
		}

		for(n=0; n < img->getNbColors(); n++) {
			colormap[n].red = colorTable[n*elementSize+0];
			colormap[n].green = colorTable[n*elementSize+1];
			colormap[n].blue = colorTable[n*elementSize+2];
			if (tableHasAlpha) {
				colormap[n].alpha = colorTable[n*elementSize+3];
			} else {
				colormap[n].alpha = 255;
			}
		}

		delete colorTable;

	} else if (format == 4) {
		img->setDepth(2);
		width = (width+1)/2*2;
		img->setBpl(width);
	} else if (format == 5) {
		img->setDepth(4);
	}

	data = new UI8[img->getDepth()*img->getWidth()*img->getHeight()];
	if (data == NULL) {
		if (colormap) delete []colormap;
		return false;
	}

	stream.next_out = data;
	stream.avail_out = img->getDepth()*img->getWidth()*img->getHeight();
	if (format != 3) status = inflateInit(&stream);

	while (1) {
		status = inflate(&stream, Z_SYNC_FLUSH) ;
		if (status == Z_STREAM_END) {
				break;
		}
		if (status != Z_OK) {
			printf("Zlib data error : %s\n", stream.msg);
			delete []data;
			return false;
		}
		stream.avail_in = 1;
	}

	inflateEnd(&stream);

	UI8*pixels = new UI8[height*width];
	if (pixels == NULL) {
		if (colormap) delete []colormap;
		delete []data;
		return false;
	}

	if (format != 3) {
		int n,c;
		unsigned char r,g,b,a;
		unsigned char *ptr;
		r = g = b = a = 0; 

		img->setNbColors(0);
		colormap = new RGBA[256];
		if (colormap == NULL) {
			delete []data;
			delete []pixels;
			return false;
		}
		memset(colormap, 0, 256 * sizeof(RGBA));
		ptr = pixels;
		
		for(n=0; n < img->getDepth()*img->getWidth()*img->getHeight(); n+=img->getDepth(),ptr++) {
			switch (format) {
				case 4:
					a = 1;
					r = (data[n] & 0x78)<<1;
					g = ((data[n] & 0x03)<<6) | (data[n+1] & 0xc0)>>2;
					b = (data[n+1] & 0x1e)<<3;
					break;
				case 5:
					a = data[n];
					// Reduce color dynamic range
					r = data[n+1]&0xe0;
					g = data[n+2]&0xe0;
					b = data[n+3]&0xe0;
					break;
			}
			for(c=0; c < img->getNbColors(); c++) {
				if (r == colormap[c].red
				&&  g == colormap[c].green
				&&  b == colormap[c].blue) {
					*ptr = c;
					break;
				}
			}
			if (c == img->getNbColors()) {
				if (img->getNbColors() == 256) continue;
				img->setNbColors(img->getNbColors()+1);
				if (img->getNbColors() == 256) {
					//printf("Colormap entries exhausted. After %d scanned pixels\n", n/4);
				}
				colormap[c].alpha = a;
				colormap[c].red   = r;
				colormap[c].green = g;
				colormap[c].blue  = b;
				*ptr = c;
			}
		}
	} else {
		memcpy(pixels, data, img->getWidth()*img->getHeight());
		if (tableHasAlpha) {
			int n;
			unsigned char *ptr, *alpha;

			UI8 *alpha_buf = new UI8[img->getWidth()*img->getHeight()];
			ptr = data;
			alpha = alpha_buf;
			for(n=0; n < img->getWidth()*img->getHeight(); n++, ptr++, alpha++) {
				*alpha = colormap[*ptr].alpha;
			}
			img->setAlphaBuf(alpha_buf);
		}
	}
	img->setColormap(colormap);	
	img->setPixels(pixels);
	delete []data;
	return true;
}

/*
 * Read JPEG image using pre-loaded Tables
 **/
bool UncompressTools::buildFromJpegAbbreviatedData(UI8* dataStream,UI8* tableStream,Image *img) {
	JSAMPROW buffer[1];
	unsigned char *ptrPix;
	int stride;
	long n;
	int status;
	MyErrorHandler jpegErrorMgr;
	struct jpeg_source_mgr jpegSourceManager;


	pthread_mutex_lock(&synchro);
	// Read header
	struct jpeg_decompress_struct jpegObject;
	readJpegTables(tableStream,jpegObject,jpegErrorMgr,jpegSourceManager);

	// Set current stream pointer to stream
	inputData = dataStream;
	if (setjmp(jpegErrorMgr.setjmp_buffer)) {
		return false;
	}

	jpeg_read_header(&jpegObject, TRUE);
	jpegObject.quantize_colors = TRUE;	// Create colormapped image
	jpeg_start_decompress(&jpegObject);

	// Set objet dimensions
	img->setHeight(jpegObject.output_height);
	img->setWidth(jpegObject.output_width);
	img->setBpl(img->getWidth());
	UI8 *pixels = new UI8[img->getHeight()*img->getWidth()];
	if (pixels == NULL) {
		jpeg_finish_decompress(&jpegObject);
		return false;
	}
	ptrPix = pixels;
	stride = jpegObject.output_width * jpegObject.output_components;
	buffer[0] = (JSAMPROW)malloc(stride);

	while (jpegObject.output_scanline < jpegObject.output_height) {
		status = jpeg_read_scanlines(&jpegObject, buffer, 1);
		memcpy(ptrPix,buffer[0],stride);
		ptrPix+= stride;
	}
	free(buffer[0]);

	RGBA *colormap = new RGBA[jpegObject.actual_number_of_colors];
	if (colormap == NULL) {
		jpeg_finish_decompress(&jpegObject);
		delete pixels;
		return false;
	}
	int nbColors = jpegObject.actual_number_of_colors;

	for(n=0; n < nbColors; n++)	{
		colormap[n].red = jpegObject.colormap[0][n];
		colormap[n].green = jpegObject.colormap[1][n];
		colormap[n].blue = jpegObject.colormap[2][n];
		colormap[n].alpha = 255;
	}

	status = jpeg_finish_decompress(&jpegObject);
	pthread_mutex_unlock(&synchro);
	img->setPixels(pixels);
	img->setColormap(colormap);	
	img->setNbColors(nbColors);
	return true;
}

/*
 * Just init JPEG object and read JPEG Tables
 **/
bool UncompressTools::readJpegTables(UI8 *stream, struct jpeg_decompress_struct &jpegObject,MyErrorHandler &jpegErrorMgr,struct jpeg_source_mgr &jpegSourceManager) {
	// Setup error handler
	jpegObject.err = jpeg_std_error(&jpegErrorMgr.pub);
	jpegErrorMgr.pub.error_exit = errorExit;

	if (setjmp(jpegErrorMgr.setjmp_buffer)) {
		jpeg_destroy_decompress(&jpegObject);
		return false;
	}
	inputData = stream;
	jpeg_create_decompress(&jpegObject);

	// Setup source manager structure
	jpegSourceManager.init_source = initSource;
	jpegSourceManager.fill_input_buffer = fillInputBuffer;
	jpegSourceManager.skip_input_data = skipInputData;
	jpegSourceManager.resync_to_restart = resyncToRestart;
	jpegSourceManager.term_source = termSource;

	// Set default source manager
	jpegObject.src = &jpegSourceManager;
	jpeg_read_header(&jpegObject, FALSE);
	return true;
}
